import os
import sys
import threading
from pprint import pprint

import pygtk
pygtk.require('2.0')
import gtk
import gobject
import hippo
import pygame

__all__ = ['PyGameCanvas']

class MockEvent(object):
    """Used to inject key-repeat events."""
   
    def __init__(self, keyval):
        self.keyval = keyval

class PyGameCanvas(gtk.EventBox):
   
    key_trans = {
        'Alt_L': pygame.K_LALT,
        'Alt_R': pygame.K_RALT,
    }
   
    mod_map = {
        pygame.K_LALT: pygame.KMOD_LALT,
        pygame.K_RALT: pygame.KMOD_RALT,
    }
   
    def __init__(self, width, height):

        # Build the main widget
        super(PyGameCanvas,self).__init__()
       
        # Build the sub-widgets
        self._align = gtk.Alignment(0.5, 0.5)
        self._evb = gtk.EventBox()
        self._socket = gtk.Socket()
       
        # Set our X event masks
        self.set_events(
            gtk.gdk.KEY_PRESS_MASK | \
            gtk.gdk.KEY_RELEASE_MASK \
        )
       
        self._evb.set_events(
            gtk.gdk.POINTER_MOTION_MASK | \
            gtk.gdk.POINTER_MOTION_HINT_MASK | \
            gtk.gdk.BUTTON_MOTION_MASK | \
            gtk.gdk.BUTTON_PRESS_MASK | \
            gtk.gdk.BUTTON_RELEASE_MASK
        )
       
        # Add internal widgets
        self._evb.set_size_request(width, height)
        self._evb.add(self._socket)
        self._socket.show()
        self._align.add(self._evb)
        self._evb.show()
        self.add(self._align)
        self._align.show()

        # Callback functions to link the event systems
        self.connect('unrealize', self._quit)
        self.connect('key_press_event', self._keydown)
        self.connect('key_release_event', self._keyup)
        self._evb.connect('button_press_event', self._mousedown)
        self._evb.connect('button_release_event', self._mouseup)
        self._evb.connect('motion-notify-event', self._mousemove)

        # <Cue Thus Spract Zarathustra>
        self.show()
        self.set_flags(gtk.CAN_FOCUS)
       
        # Internal data
        self.__stopped = False
        self.__keystate = [0] * 323
        self.__mouse_pos = (0,0)
        self.__repeat = (None, None)
        self.__held = set()
        self.__held_time_left = {}
        self.__held_last_time = {}
        self.__tick_id = None
       
    def connect_game(self, app):
        # Setup the embedding
        os.environ['SDL_WINDOWID'] = str(self._socket.get_id())
        #print 'Socket ID=%s'%os.environ['SDL_WINDOWID']
        pygame.init()
       
        # Hijack some function
        pygame.key.get_pressed = self._get_pressed
        pygame.key.set_repeat = self._set_repeat
       
        # Load the modules
        # NOTE: This is delayed because pygame.init() must come after the embedding is up
        if ':' not in app:
            app += ':main'
        mod_name, fn_name = app.split(':')
        mod = __import__(mod_name, globals(), locals(), [])
        fn = getattr(mod, fn_name)
       
        # Start Pygame
        self.__thread = threading.Thread(target=self._start, args=[fn])
        self.__thread.start()
       
    def _start(self, fn):
        fn()
        if not self.__stopped:
            gtk.main_quit()
       
    def _quit(self, data=None):
        print 'Quitting pygame'
        self.__stopped = True
        pygame.event.post(pygame.event.Event(pygame.QUIT))
        self.__thread.join()
        print 'Gone'

    def _keydown(self, widget, event):
        key = event.keyval
        if key in self.__held:
            return True
        else:
            if self.__repeat[0] is not None:
                self.__held_last_time[key] = pygame.time.get_ticks()
                self.__held_time_left[key] = self.__repeat[0]
            self.__held.add(key)
           
        return self._keyevent(widget, event, pygame.KEYDOWN)
       
    def _keyup(self, widget, event):
        key = event.keyval
        if self.__repeat[0] is not None:
            if key in self.__held:
                # This is possibly false if set_repeat() is called with a key held
                del self.__held_time_left[key]
                del self.__held_last_time[key]
        self.__held.discard(key)

        return self._keyevent(widget, event, pygame.KEYUP)
       
    def _keyevent(self, widget, event, type):
        key = gtk.gdk.keyval_name(event.keyval)
        keycode = None
        if key in self.key_trans:
            keycode = self.key_trans[key]
        elif hasattr(pygame, 'K_'+key.upper()):
            keycode = getattr(pygame, 'K_'+key.upper())
        elif hasattr(pygame, 'K_'+key.lower()):
            keycode = getattr(pygame, 'K_'+key.lower())
        else:
            print 'Key %s unrecognized'%key
           
        if keycode is not None:
            self.__keystate[keycode] = type == pygame.KEYDOWN
            ukey = gtk.gdk.keyval_to_unicode(event.keyval)
            mod = 0
            for key_val, mod_val in self.mod_map.iteritems():
                mod |= self.__keystate[key_val] and mod_val
           
            evt = pygame.event.Event(type, key=keycode, unicode=ukey)
            pygame.event.post(evt)
        return True

    def _get_pressed(self):
        return self.__keystate

    def _mousedown(self, widget, event):
        return self._mouseevent(widget, event, pygame.MOUSEBUTTONDOWN)

    def _mouseup(self, widget, event):
        return self._mouseevent(widget, event, pygame.MOUSEBUTTONUP)
       
    def _mouseevent(self, widget, event, type):
        evt = pygame.event.Event(type,
                                             button=event.button,
                                             x=event.x,
                                             y=event.y)
        pygame.event.post(evt)
        return True
       
    def _mousemove(self, widget, event):
        # From http://www.learningpython.com/2006/07/25/writing-a-custom-widget-using-pygtk/
        # if this is a hint, then let's get all the necessary
        # information, if not it's all we need.
        if event.is_hint:
                x, y, state = event.window.get_pointer()
        else:
                x = event.x
                y = event.y
                state = event.state
       
        rel = (self.__mouse_pos[0] - x,
                self.__mouse_pos[1] - y)
        self.__mouse_pos = (x, y)
       
        buttons = (
            state & gtk.gdk.BUTTON1_MASK and 1 or 0,
            state & gtk.gdk.BUTTON2_MASK and 1 or 0,
            state & gtk.gdk.BUTTON3_MASK and 1 or 0,
        )
       
        evt = pygame.event.Event(pygame.MOUSEMOTION,
                                             pos=self.__mouse_pos,
                                             rel=rel,
                                             buttons=buttons)
        pygame.event.post(evt)
        return True
       
    def _tick(self):
        cur_time = pygame.time.get_ticks()
        for key in self.__held:
            delta = cur_time - self.__held_last_time[key]
            self.__held_last_time[key] = cur_time
           
            self.__held_time_left[key] -= delta
            if self.__held_time_left[key] <= 0:
                self.__held_time_left[key] = self.__repeat[1]
                self._keyevent(None, MockEvent(key), pygame.KEYDOWN)
               
        return True
       
    def _set_repeat(self, delay=None, interval=None):
        if delay is not None and self.__repeat[0] is None:
            self.__tick_id = gobject.timeout_add(10, self._tick)
        elif delay is None and self.__repeat[0] is not None:
            gobject.source_remove(self.__tick_id)
        self.__repeat = (delay, interval)